from tableau_api_lib.api_endpoints import BaseEndpoint


class TasksEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 get_refresh_tasks=False,
                 get_flow_run_tasks=False,
                 get_refresh_task=False,
                 get_flow_run_task=False,
                 run_refresh_task=False,
                 run_flow_task=False,
                 task_id=None,
                 flow_id=None,
                 query_schedule_refresh_tasks=False,
                 schedule_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API task methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool get_refresh_tasks: True if getting all refresh tasks, False otherwise.
        :param bool get_refresh_task: True if getting a specific refresh task, False otherwise.
        :param bool run_refresh_task: True if running a specific refresh task, False otherwise.
        :param bool run_flow_task: True if running a specific flow task, False otherwise.
        :param str task_id: the task ID.
        :param str flow_id: the flow ID.
        :param bool query_schedule_refresh_tasks: True if querying all refresh tasks, False otherwise.
        :param str schedule_id: the schedule ID.
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint.
        """

        super().__init__(ts_connection)
        self._get_refresh_tasks = get_refresh_tasks
        self._get_flow_run_tasks = get_flow_run_tasks
        self._get_refresh_task = get_refresh_task
        self._get_flow_run_task = get_flow_run_task
        self._run_refresh_task = run_refresh_task
        self._run_flow_task = run_flow_task
        self._task_id = task_id
        self._flow_id = flow_id
        self._query_schedule_refresh_tasks = query_schedule_refresh_tasks
        self._schedule_id = schedule_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._get_refresh_tasks,
            self._get_flow_run_tasks,
            self._get_refresh_task,
            self._get_flow_run_task,
            self._run_refresh_task,
            self._run_flow_task,
            self._query_schedule_refresh_tasks
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_task_url(self):
        return "{0}/api/{1}/sites/{2}/tasks".format(self._connection.server,
                                                    self._connection.api_version,
                                                    self._connection.site_id)

    @property
    def base_schedules_url(self):
        return "{0}/api/{1}/sites/{2}/schedules".format(self._connection.server,
                                                        self._connection.api_version,
                                                        self._connection.site_id)

    @property
    def base_extract_refresh_url(self):
        return "{0}/extractRefreshes".format(self.base_task_url)

    @property
    def base_extract_refresh_id_url(self):
        if self._run_refresh_task:
            return "{0}/{1}/runNow".format(self.base_extract_refresh_url,
                                           self._task_id)
        else:
            return "{0}/{1}".format(self.base_extract_refresh_url,
                                    self._task_id)

    @property
    def base_query_extract_refresh_url(self):
        return "{0}/{1}/extracts".format(self.base_schedules_url,
                                         self._schedule_id)

    @property
    def base_flow_run_tasks_url(self):
        return "{0}/runFlow".format(self.base_task_url)

    @property
    def base_flow_run_task_id_url(self):
        if self._run_flow_task:
            return "{0}/{1}/runNow".format(self.base_flow_run_tasks_url,
                                           self._task_id)
        else:
            return "{0}/{1}".format(self.base_flow_run_tasks_url,
                                    self._task_id)

    def get_endpoint(self):
        url = None
        if self._get_refresh_tasks:
            url = self.base_extract_refresh_url
        elif self._get_refresh_task or self._run_refresh_task:
            url = self.base_extract_refresh_id_url
        elif self._query_schedule_refresh_tasks:
            url = self.base_query_extract_refresh_url
        elif self._get_flow_run_tasks:
            url = self.base_flow_run_tasks_url
        elif self._get_flow_run_task:
            url = self.base_flow_run_task_id_url
        elif self._run_flow_task:
            url = self.base_flow_run_task_id_url
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from .users import get_users_dataframe
from .groups import get_groups_dataframe, get_group_users_dataframe
from .datasources import get_datasources_dataframe, get_datasource_connections_dataframe
from .workbooks import get_workbooks_dataframe, get_views_dataframe, get_workbook_connections_dataframe, \
    get_embedded_datasources_dataframe
from .sites import get_sites_dataframe, get_active_site_content_url, get_active_site_name, get_active_site_id
from .projects import get_projects_dataframe
from .schedules import get_schedules_dataframe
from .favorites import get_user_favorites_dataframe
from .flows import get_flows_dataframe



from tableau_api_lib.api_endpoints import BaseEndpoint


class PermissionsEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 add_object_permissions=False,
                 query_object_permissions=False,
                 delete_object_permissions=False,
                 object_type=None,
                 object_id=None,
                 query_default_project_permissions=False,
                 add_default_project_permissions=False,
                 delete_default_project_permissions=False,
                 project_permissions_object=None,
                 delete_permissions_object=None,
                 delete_permissions_object_id=None,
                 capability_name=None,
                 capability_mode=None,
                 project_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API permissions methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool add_object_permissions: True if adding object permissions, False otherwise
        :param bool query_object_permissions: True if querying object permissions, False otherwise
        :param bool delete_object_permissions: True if deleting object permissions, False otherwise
        :param str object_type: the Tableau object type (workbook, etc.)
        :param str object_id: the Tableau object ID
        :param bool query_default_project_permissions: True if querying default project permissions, False otherwise
        :param bool add_default_project_permissions: True if adding default project permissions, False otherwise
        :param bool delete_default_project_permissions: True if deleting default project permissions, False otherwise
        :param str project_permissions_object: the project permissions object (workbook, etc.)
        :param str delete_permissions_object: the permissions object to delete (user, group, etc.)
        :param str delete_permissions_object_id: the ID of the permissions object to delete (user ID, group ID, etc.)
        :param str capability_name: the permissions capability being targeted (only use when deleting)
        :param str capability_mode: set this value to 'Allow' when you want to delete a permission allowing a
        capability; set this value to 'Deny' when you want to delete a permission denying a capability
        :param str project_id: the project ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._add_object_permissions = add_object_permissions
        self._query_object_permissions = query_object_permissions
        self._delete_object_permissions = delete_object_permissions
        self._object_type = self._enforce_plurals(object_type,
                                                  self.valid_project_permissions_objects)
        self._object_id = object_id
        self._query_default_project_permissions = query_default_project_permissions
        self._add_default_project_permissions = add_default_project_permissions
        self._delete_default_project_permissions = delete_default_project_permissions
        self._project_permissions_object = self._enforce_plurals(project_permissions_object,
                                                                 self.valid_project_permissions_objects)
        self._delete_permissions_object = self._enforce_plurals(delete_permissions_object,
                                                                self.valid_project_delete_permissions_objects)
        self._delete_permissions_object_id = delete_permissions_object_id
        self._capability_name = capability_name
        self._capability_mode = capability_mode
        self._project_id = project_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._add_object_permissions,
            self._query_object_permissions,
            self._delete_object_permissions,
            self._query_default_project_permissions,
            self._add_default_project_permissions,
            self._delete_default_project_permissions
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def valid_project_permissions_objects(self):
        return [
            'datasources',
            'datasource',
            'workbooks',
            'workbook',
            'flows',
            'flow',
            'views',
            'view',
            'projects',
            'project'
        ]

    @property
    def valid_project_delete_permissions_objects(self):
        return [
            'users',
            'user',
            'groups',
            'group'
        ]

    def _enforce_plurals(self, permissions_object, valid_permissions_objects):
        print('permissions_object: ', permissions_object)
        print('valid_permissions_objects: ', valid_permissions_objects)
        if permissions_object:
            if permissions_object in valid_permissions_objects:
                if permissions_object[-1] == 's':
                    return permissions_object
                else:
                    return permissions_object + 's'
            else:
                self._invalid_parameter_exception()

    @property
    def base_permissions_url(self):
        return "{0}/api/{1}/sites/{2}".format(self._connection.server,
                                              self._connection.api_version,
                                              self._connection.site_id)

    @property
    def base_object_permissions_url(self):
        return "{0}/{1}/{2}/permissions".format(self.base_permissions_url,
                                                self._object_type,
                                                self._object_id)

    @property
    def base_query_default_permissions_url(self):
        return "{0}/projects/{1}/default-permissions/{2}".format(self.base_permissions_url,
                                                                 self._project_id,
                                                                 self._project_permissions_object)

    @property
    def base_delete_permission_url(self):
        return "{0}/{1}/{2}/{3}/{4}".format(self.base_object_permissions_url,
                                            self._delete_permissions_object,
                                            self._delete_permissions_object_id,
                                            self._capability_name,
                                            self._capability_mode)

    @property
    def base_delete_default_permissions_url(self):
        return "{0}/{1}/{2}/{3}/{4}".format(self.base_query_default_permissions_url,
                                            self._delete_permissions_object,
                                            self._delete_permissions_object_id,
                                            self._capability_name,
                                            self._capability_mode)

    def get_endpoint(self):
        url = None
        if self._add_object_permissions:
            url = self.base_object_permissions_url
        elif self._query_object_permissions:
            url = self.base_object_permissions_url
        elif self._delete_object_permissions:
            url = self.base_delete_permission_url
        elif self._delete_default_project_permissions:
            url = self.base_delete_default_permissions_url
        elif self._query_default_project_permissions:
            url = self.base_query_default_permissions_url
        elif self._add_default_project_permissions:
            url = self.base_query_default_permissions_url
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from tableau_api_lib.api_requests import BaseRequest


class AddUserToSiteRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding users to sites.
    :param class ts_connection: the Tableau Server connection object
    :param str user_name: the username for the user being added
    :param str site_role: the site role to assign to the added user
    :param str auth_setting: the auth setting to assign to the added user
    """
    def __init__(self,
                 ts_connection,
                 user_name,
                 site_role='Viewer',
                 auth_setting=None):
        super().__init__(ts_connection)
        self._user_name = user_name if isinstance(user_name, str) else None
        self._site_role = site_role
        self._auth_setting = auth_setting if isinstance(auth_setting, str) else None
        self._validate_site_role()
        self.base_add_user_request()

    @property
    def required_user_param_keys(self):
        return [
            'name',
            'siteRole'
        ]

    @property
    def optional_user_param_keys(self):
        return ['authSetting']

    @property
    def required_user_param_values(self):
        return [
            self._user_name,
            self._site_role
        ]

    @property
    def optional_user_param_values(self):
        return [self._auth_setting]

    @property
    def valid_site_roles(self):
        return [
            'Creator',
            'creator',
            'Explorer',
            'explorer',
            'ExplorerCanPublish',
            'explorercanpublish',
            'siteAdministratorExplorer',
            'siteadministratorexplorer',
            'SiteAdministratorCreator',
            'siteadministratorcreator',
            'ServerAdministrator',
            'serverAdministrator',
            'Unlicensed',
            'unlicensed',
            'Viewer',
            'viewer'
        ]

    def _validate_site_role(self):
        valid = True
        if not(self._site_role in self.valid_site_roles):
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    def base_add_user_request(self):
        self._request_body.update({'user': {}})
        self._request_body['user'].update(
            self._get_parameters_dict(self.required_user_param_keys,
                                      self.required_user_param_values))
        return self._request_body

    def modified_add_user_request(self):
        if any(self.optional_user_param_values):
            self._request_body['user'].update(
                self._get_parameters_dict(self.optional_user_param_keys,
                                          self.optional_user_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_add_user_request()



import numpy as np
from tableau_api_lib.api_requests import BaseRequest


class UpdateUserRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating databases.
    :param class ts_connection: the Tableau Server connection object
    :param str new_full_name: (optional) the new name for the user. Users can change names without affecting the groups
    they belong to. Tableau Server only; not available in Tableau Online
    :param str new_email: (optional) the new email address for the user; Tableau Server only; not available in Tableau
    Online
    :param str new_password: (optional) the new password for the user; Tableau Server only; not available in Tableau
    Online
    :param str new_site_role: (optional) the new site role. Valid role names are
    [ServerAdministratorExplorer, ExplorerCanPublish, SiteAdministrator, or Unlicensed]; see the Tableau Server REST API
    documentation for further details.
    :param str new_auth_setting: (optional) the new authentication type for the user. You can assign the following
    values for this attribute: SAML (the user signs in using SAML) or ServerDefault (the user signs in using the
    authentication method that's set for the server); see the Tableau Server REST API documentation for further details.
    """
    def __init__(self,
                 ts_connection,
                 new_full_name=None,
                 new_email=None,
                 new_password=None,
                 new_site_role=None,
                 new_auth_setting=None):

        super().__init__(ts_connection)
        self._new_full_name = new_full_name if isinstance(new_full_name, str) else None
        self._new_email = new_email if isinstance(new_email, str) else None
        self._new_password = new_password
        self._new_site_role = new_site_role
        self._new_auth_setting = new_auth_setting if isinstance(new_auth_setting, str) else None
        self.base_update_user_request()

    @property
    def optional_user_param_keys(self):
        return [
            'fullName',
            'email',
            'password',
            'siteRole',
            'authSetting'
        ]

    @property
    def optional_user_param_values(self):
        return [
            self._new_full_name,
            self._new_email,
            self._new_password,
            self._new_site_role,
            self._new_auth_setting
        ]

    def base_update_user_request(self):
        self._request_body.update({'user': {}})
        return self._request_body

    def modified_update_user_request(self):
        if any(self.optional_user_param_values):
            self._request_body['user'].update(
                self._get_parameters_dict(self.optional_user_param_keys,
                                          self.optional_user_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_update_user_request()



from tableau_api_lib.api_requests import BaseRequest


class UpdateColumnRequest(BaseRequest):
    def __init__(self,
                 ts_connection,
                 new_description_value=None):
        """
        Builds the request body for Tableau Server REST API requests updating database columns.
        :param class ts_connection: the Tableau Server connection object
        :param str new_description_value: (optional) the new description for the column
        """

        super().__init__(ts_connection)
        self._new_description_value = new_description_value
        self._request_body = {'column': {}}

    @property
    def optional_param_keys(self):
        return ['description']

    @property
    def optional_param_values(self):
        return [self._new_description_value]

    def base_update_column_request(self):
        self._request_body['column'].update(self._get_parameters_dict(self.optional_param_keys,
                                                                      self.optional_param_values))
        return self._request_body

    def get_request(self):
        return self.base_update_column_request()



from tableau_api_lib.exceptions import InvalidParameterException


class BaseRequest:
    """
    The base request from which all other API requests are built.
    :param class ts_connection: the Tableau Server connection object
    """
    def __init__(self,
                 ts_connection):

        self._connection = ts_connection
        self._request_body = {}

    @staticmethod
    def _get_parameters_dict(param_keys, param_values):
        params_dict = {}
        for i, key in enumerate(param_keys):
            if param_values[i] or param_values[i] is False:
                params_dict.update({key: param_values[i]})
        return params_dict

    def _invalid_parameter_exception(self):
        raise InvalidParameterException(class_name=self.__class__.__name__,
                                        parameters=vars(self))



from tableau_api_lib.api_endpoints import BaseEndpoint


class ViewEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_views=False,
                 query_view=False,
                 view_id=None,
                 add_tags=False,
                 delete_tag=False,
                 tag_name=None,
                 query_view_pdf=False,
                 query_view_image=False,
                 query_view_data=False,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API view methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_views: True if querying all views, False otherwise
        :param bool query_view: True if querying a specific view, False otherwise
        :param str view_id: the view ID
        :param bool add_tags: True if adding tags, False otherwise
        :param bool delete_tag: True if deleting a specific tag, False otherwise
        :param str tag_name: the name of the tag
        :param bool query_view_pdf: True if querying a specific view's PDF, False otherwise
        :param bool query_view_image: True if querying a specific view's image, False otherwise
        :param bool query_view_data: True if querying a specific view's data, False otherwise
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._view_id = view_id
        self._add_tags = add_tags
        self._delete_tag = delete_tag
        self._tag_name = tag_name
        self._query_view = query_view
        self._query_views = query_views
        self._query_view_pdf = query_view_pdf
        self._query_view_image = query_view_image
        self._query_view_data = query_view_data
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._add_tags,
            self._delete_tag,
            self._query_view,
            self._query_views,
            self._query_view_pdf,
            self._query_view_image,
            self._query_view_data
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_view_url(self):
        return "{0}/api/{1}/sites/{2}/views".format(self._connection.server,
                                                    self._connection.api_version,
                                                    self._connection.site_id)

    @property
    def base_view_id_url(self):
        return "{0}/{1}".format(self.base_view_url,
                                self._view_id)

    @property
    def base_view_tags_url(self):
        return "{0}/tags".format(self.base_view_id_url)

    @property
    def base_delete_view_tag_url(self):
        return "{0}/{1}".format(self.base_view_tags_url,
                                self._tag_name)

    @property
    def base_query_view_pdf_url(self):
        return "{0}/{1}/pdf".format(self.base_view_url,
                                    self._view_id)

    @property
    def base_query_view_image_url(self):
        return "{0}/{1}/image".format(self.base_view_url,
                                      self._view_id)

    @property
    def base_query_view_data_url(self):
        return "{0}/data".format(self.base_view_id_url)

    def get_endpoint(self):
        url = None
        if self._view_id:
            if self._query_view:
                url = self.base_view_id_url
            elif self._add_tags:
                url = self.base_view_tags_url
            elif self._delete_tag and self._tag_name:
                url = self.base_delete_view_tag_url
            elif self._query_view_pdf:
                url = self.base_query_view_pdf_url
            elif self._query_view_image:
                url = self.base_query_view_image_url
            elif self._query_view_data:
                url = self.base_query_view_data_url
            else:
                self._invalid_parameter_exception()
        else:
            url = self.base_view_url

        return self._append_url_parameters(url)



from tableau_api_lib.api_endpoints import BaseEndpoint


class SubscriptionsEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 create_subscription=False,
                 query_subscriptions=False,
                 query_subscription=False,
                 update_subscription=False,
                 delete_subscription=False,
                 subscription_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API subscription methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool create_subscription: True if creating a subscription, False otherwise
        :param bool query_subscriptions: True if querying all subscriptions, False otherwise
        :param bool query_subscription: True if querying a specific subscription, False otherwise
        :param bool update_subscription: True if updating a specific subscription, False otherwise
        :param bool delete_subscription: True if deleting a specific subscription, False otherwise
        :param str subscription_id: the subscription ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._create_subscription = create_subscription
        self._query_subscriptions = query_subscriptions
        self._query_subscription = query_subscription
        self._update_subscription = update_subscription
        self._delete_subscription = delete_subscription
        self._subscription_id = subscription_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._create_subscription,
            self._query_subscriptions,
            self._query_subscription,
            self._update_subscription,
            self._delete_subscription
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_subscription_url(self):
        return "{0}/api/{1}/sites/{2}/subscriptions".format(self._connection.server,
                                                            self._connection.api_version,
                                                            self._connection.site_id)

    @property
    def base_subscription_id_url(self):
        return "{0}/{1}".format(self.base_subscription_url,
                                self._subscription_id)

    def get_endpoint(self):
        url = None
        if self._create_subscription:
            url = self.base_subscription_url
        elif self._query_subscriptions and not self._subscription_id:
            url = self.base_subscription_url
        elif self._subscription_id:
            if self._query_subscription or self._update_subscription or self._delete_subscription:
                url = self.base_subscription_id_url
            else:
                self._invalid_parameter_exception()
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from tableau_api_lib.api_requests import BaseRequest


class UpdateGroupRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating groups.
    :param class ts_connection: the Tableau Server connection object
    :param str new_group_name: the new name for the group
    :param str active_directory_group_name: the name of the Active Directory group to synchronize with
    :param str active_directory_domain_name: the domain for the Active Directory group
    :param str default_site_role: the site role to assign to users that are synchronized with Active Directory. You can
    assign the following roles:
    [Creator, Explorer, ExplorerCanPublish, SiteAdministratorExplorer, SiteAdministratorCreator, Unlicensed, or Viewer]
    """
    def __init__(self,
                 ts_connection,
                 new_group_name,
                 active_directory_group_name=None,
                 active_directory_domain_name=None,
                 default_site_role=None):
        super().__init__(ts_connection)
        self._new_group_name = new_group_name
        self._active_directory_group_name = active_directory_group_name
        self._active_directory_domain_name = active_directory_domain_name
        self._default_site_role = default_site_role
        self.base_update_group_request()

    @property
    def required_group_param_keys(self):
        return ['name']

    @property
    def optional_import_param_keys(self):
        return [
            'source',
            'domainName',
            'siteRole'
        ]

    @property
    def required_group_param_values(self):
        return [self._new_group_name]

    @property
    def optional_import_param_values(self):
        return [
            self._active_directory_group_name,
            self._active_directory_domain_name,
            self._default_site_role
        ]

    def base_update_group_request(self):
        self._request_body.update({'group': {}})
        self._request_body['group'].update(
            self._get_parameters_dict(self.required_group_param_keys,
                                      self.required_group_param_values))
        return self._request_body

    def modified_update_group_request(self):
        if any(self.optional_import_param_values):
            self._request_body['group'].update({'import': {}})
            self._request_body['group']['import'].update(
                self._get_parameters_dict(self.optional_import_param_keys,
                                          self.optional_import_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_update_group_request()



from tableau_api_lib.api_requests import BaseRequest


class SignInRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests when signing in.
    :param class ts_connection: the Tableau Server connection object
    :param str username: the username credentials for signing in to Tableau Server
    :param str password: the password credentials for signing in to Tableau Server
    :param str personal_access_token_name: the personal access token name for signing in to Tableau Server
    :param str personal_access_token_secret: the personal access token secret for signing in to Tableau Server
    :param str user_to_impersonate: if impersonating another user, set this value with their user ID
    """
    def __init__(self,
                 ts_connection,
                 auth_method,
                 username,
                 password,
                 personal_access_token_name,
                 personal_access_token_secret,
                 user_to_impersonate=None):

        super().__init__(ts_connection)
        self._auth_method = auth_method
        self._username = username
        self._password = password
        self._personal_access_token_name = personal_access_token_name
        self._personal_access_token_secret = personal_access_token_secret
        self._user_to_impersonate = user_to_impersonate
        self.base_signin_request()

    def base_signin_request(self):
        if self._auth_method == "user_and_password":
            self._request_body.update({
                "credentials": {
                    "name": self._username,
                    "password": self._password,
                    "site": {"contentUrl": self._connection.site_url}
                }
            })
        elif self._auth_method == "personal_access_token":
            self._request_body.update({
                "credentials": {
                    "personalAccessTokenName": self._personal_access_token_name,
                    "personalAccessTokenSecret": self._personal_access_token_secret,
                    "site": {"contentUrl": self._connection.site_url}
                }
            })
        return self._request_body

    def modified_signin_request(self):
        self._request_body['credentials'].update({
            "user": {
                "id": self._user_to_impersonate
            }
        })
        return self._request_body

    def get_request(self):
        if self._username and self._password and not self._user_to_impersonate:
            return self._request_body
        elif self._username and self._password and self._user_to_impersonate:
            return self.modified_signin_request()
        elif self._personal_access_token_name and self._personal_access_token_secret:
            return self.modified_signin_request()
        else:
            self._invalid_parameter_exception()



import os
import shutil
import zipfile
from tableau_api_lib import TableauServerConnection
from tableau_api_lib.utils.common import get_server_netloc


def remap_xml_references(file, conn_source, conn_target) -> str:
    """
    Replaces any reference to the source server with a reference to the target server.
    :param str file: the file contents as a string of text
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_target: the target Tableau Server connection
    :return: file contents as a string of text
    """
    file = file.replace("xml:base='{}'".format(conn_source.server),
                        "xml:base='{}'".format(conn_target.server))
    file = file.replace("xml:base=&apos;{}".format(conn_source.server),
                        "xml:base=&apos;{}".format(conn_target.server))
    file = file.replace("path='/t/{}/".format(conn_source.site_url),
                        "path='/t/{}/".format(conn_target.site_url))
    file = file.replace("path=&apos;/t/{}/".format(conn_source.site_url),
                        "path=&apos;/t/{}/".format(conn_target.site_url))
    file = file.replace("site='{}'".format(conn_source.site_url),
                        "site='{}'".format(conn_target.site_url))
    file = file.replace("site=&apos;{}".format(conn_source.site_url),
                        "site=&apos;{}".format(conn_target.site_url))
    file = file.replace("server='{}'".format(get_server_netloc(conn_source.server)),
                        "server='{}'".format(get_server_netloc(conn_target.server)))
    file = file.replace("server=&apos;{}".format(get_server_netloc(conn_source.server)),
                        "server=&apos;{}".format(get_server_netloc(conn_target.server)))
    return file


def copy_dir_to_zip(path, zip_file_obj) -> None:
    """
    Copies any extracted contents (excluding .tds or .twb) to the new zipfile object.
    :param str path: the path to the desired extracted files and directories
    :param zipfile.ZipFile zip_file_obj: the destination zipfile object
    :return: None
    """
    for root, dirs, files in os.walk(path):
        for file in files:
            filename, file_extension = os.path.splitext(file)
            if file_extension not in ['.twb', '.tds']:
                # zip_file_obj.write(os.path.join(root, file))
                zip_file_obj.write(os.path.join(root, file), arcname=os.path.join('Data/Extracts', file))


def replace_unzipped_xml_file(file_path, conn_source, conn_target, extraction_dir_path) -> None:
    """
    Replace a .twb or .tds file, which was added as-is and not zipped.
    :param str file_path:
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_target: the target Tableau Server connection
    :param str extraction_dir_path: path to the directory where the replacement file will be written
    :return: None
    """
    with open(file_path, 'r', encoding='utf-8') as original_file:
        file_contents = original_file.read()
        file_contents = remap_xml_references(file_contents, conn_source, conn_target)
    with open(extraction_dir_path + '/' + os.path.basename(file_path), 'w', encoding='utf-8') as new_file:
        new_file.write(file_contents)


def replace_zipped_xml_file(conn_source, conn_target, zip_file, tableau_file_base, extraction_dir_path) -> None:
    """
    Replace a .twbx or .tdsx file, which was downloaded as a zipped file.
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_target: the target Tableau Server connection
    :param zipfile.ZipFile zip_file: the zipfile object
    :param str tableau_file_base: the Tableau file name
    :param str extraction_dir_path: path to the directory where the replacement file will be written
    :return: None
    """
    with zip_file.open(tableau_file_base, 'r') as file:
        file_contents = str(file.read(), 'utf-8')
        file_contents = remap_xml_references(file_contents, conn_source, conn_target)
    with open(extraction_dir_path + '/' + tableau_file_base, 'w', encoding='utf-8') as new_file:
        new_file.write(file_contents)


def generate_tableau_zipfile(extraction_dir_path, destination_dir_path, zip_file_name, xml_file_name) -> None:
    """
    Generate a Tableau .twbx or .tdsx file.
    :param str extraction_dir_path: path to the directory where the zipped file was extracted
    :param str destination_dir_path: path to the directory where the generated zipped file will be written
    :param str zip_file_name: the name of the zipped file
    :param str xml_file_name: the name of the .twb or .tds file within the zipped file
    :return: None
    """
    with zipfile.ZipFile(destination_dir_path + '/' + zip_file_name, 'w', zipfile.ZIP_DEFLATED) as modified_zipfile:
        modified_zipfile.write(extraction_dir_path + '/' + xml_file_name, xml_file_name)
        if 'Data' in os.listdir(extraction_dir_path):
            copy_dir_to_zip(path=extraction_dir_path, zip_file_obj=modified_zipfile)


def get_tableau_filenames(origin_path, extract_dir_path) -> (str, str):
    """
    Get the .twb or .tds file name from the directory where a zipped file was extracted.
    :param str origin_path: the origin path
    :param str extract_dir_path: the path to the directory where the zipped file was extracted.
    :return: str, str
    """
    origin_base = os.path.basename(origin_path)
    extracted_files = os.listdir(extract_dir_path)
    try:
        xml_file_base = [file for file in extracted_files if os.path.splitext(file)[1] in ['.tds', '.twb']].pop()
    except IndexError:
        raise Exception(f"No .tds or .twb files were discovered in the directory '{extract_dir_path}'.")
    return origin_base, xml_file_base


def modify_tableau_zipfile(zipfile_path, conn_source, conn_target, extraction_dir_path, destination_dir_path) -> None:
    """
    Modify a .twbx or .tdsx file and replace source connection references with target connection references.
    :param str zipfile_path: path to the zipped file that will be modified
    :param TableauServerConnection conn_source: the source Tableau Server connection
    :param TableauServerConnection conn_target: the target Tableau Server connection
    :param str extraction_dir_path: path to the directory where the zipped file has been extracted
    :param str destination_dir_path: path to the directory where the modified zipped file will be written
    :return: None
    """
    with zipfile.ZipFile(file=zipfile_path) as zip_file:
        print(f"extracting contents of file '{zipfile_path}'...")
        zip_file.extractall(path=extraction_dir_path)
        zip_file_base, xml_file_base = get_tableau_filenames(zipfile_path, extraction_dir_path)
        print(f"discovered a file to be modified: '{xml_file_base}'.")
        replace_zipped_xml_file(conn_source, conn_target, zip_file, xml_file_base, extraction_dir_path)
        print(f"successfully modified file '{xml_file_base}'.")
    generate_tableau_zipfile(extraction_dir_path, destination_dir_path, zip_file_base, xml_file_base)
    print(f"created the modified zipped file at '{extraction_dir_path}/{zip_file_base}'.")


def delete_temp_files(temp_dir_path=None) -> None:
    """
    Delete all directories and files nested within the specified directory path.
    :param str temp_dir_path: path to the directory whose nested contents will be deleted; the named directory persists
    :return: None
    """
    temp_dir_path = temp_dir_path or os.getcwd() + '/temp'
    print(f"deleting all nested content within dir '{temp_dir_path}'...")
    for root, dirs, files in os.walk(temp_dir_path):
        for file in files:
            print(f"removing file '{root}/{file}'")
            os.remove(os.path.join(root, file))
        for folder in dirs:
            print(f"removing dir '{root}/{folder}'")
            shutil.rmtree(os.path.join(root, folder))


def create_temp_dirs(temp_dir_path=None) -> None:
    temp_dir_path = temp_dir_path or os.getcwd() + '/temp'
    print(f"creating temp directory at '{temp_dir_path}'")
    try:
        os.makedirs(temp_dir_path, exist_ok=True)
        for content_type in ['datasources', 'workbooks', 'flows']:
            create_temp_dir_branches(temp_dir_path, content_type)
        print(f"creating dir '{temp_dir_path}/extracted'")
        os.makedirs(f"{temp_dir_path}/extracted")
    except FileExistsError:
        raise Exception(f"The directory '{temp_dir_path}' already exists and contains content.")


def create_temp_dir_branches(path, content_type):
    valid_content_types = ['datasources', 'workbooks', 'flows']
    if content_type not in valid_content_types:
        raise ValueError("Valid content types include 'datasources', 'workbooks', and 'flows'.")
    print(f"creating dir '{path}/{content_type}'")
    os.makedirs(f"{path}/{content_type}")
    print(f"creating dir '{path}/{content_type}/source'")
    os.makedirs(f"{path}/{content_type}/source")
    print(f"creating dir '{path}/{content_type}/target'")
    os.makedirs(f"{path}/{content_type}/target")



from tableau_api_lib.api_endpoints import BaseEndpoint


class GroupEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_groups=False,
                 group_id=None,
                 update_group=False,
                 delete_group=False,
                 create_group=False,
                 get_users=False,
                 add_user=False,
                 remove_user=False,
                 user_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API group methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_groups: True if querying all groups on a specific site, False otherwise
        :param str group_id: the group ID
        :param bool update_group: True if updating a specific group's information, False otherwise
        :param bool delete_group: True if deleting a specific group, False otherwise
        :param bool get_users: True if querying all users in a specific group, False otherwise
        :param bool add_user: True if adding a user to a specific group, False otherwise
        :param bool remove_user: True if removing a user from a specific group, False otherwise
        :param str user_id: the user ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._query_groups = query_groups
        self._group_id = group_id
        self._update_group = update_group
        self._delete_group = delete_group
        self._create_group = create_group
        self._get_users = get_users
        self._add_user = add_user
        self._remove_user = remove_user
        self._user_id = user_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_groups,
            self._update_group,
            self._delete_group,
            self._create_group,
            self._get_users,
            self._add_user,
            self._remove_user
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_group_url(self):
        return "{0}/api/{1}/sites/{2}/groups".format(self._connection.server,
                                                     self._connection.api_version,
                                                     self._connection.site_id)

    @property
    def base_group_id_url(self):
        return "{0}/{1}".format(self.base_group_url, self._group_id)

    @property
    def base_group_user_url(self):
        return "{0}/users".format(self.base_group_id_url)

    @property
    def base_group_user_id_url(self):
        return "{0}/{1}".format(self.base_group_user_url,
                                self._user_id)

    def get_endpoint(self):
        url = None
        if self._group_id:
            if self._update_group:
                url = self.base_group_id_url
            elif self._delete_group:
                url = self.base_group_id_url
            elif self._get_users or self._add_user:
                url = self.base_group_user_url
            elif self._remove_user and self._user_id:
                url = self.base_group_user_id_url
            else:
                self._invalid_parameter_exception()
        else:
            url = self.base_group_url

        return self._append_url_parameters(url)



"""
Helper functions for querying REST API data for datasources
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_datasource_fields(conn) -> list:
    """
    Queries all available datasource fields from Tableau Server.
    :param class conn: the Tableau Server connection
    :return: list
    """
    all_datasources = extract_pages(conn.query_data_sources, parameter_dict={'fields': 'fields=_all_'})
    return all_datasources


def get_datasources_dataframe(conn, datasource_names=None) -> pd.DataFrame:
    """
    Returns a Pandas DataFrame of all available Tableau Server datasource fields.
    :param class conn: the Tableau Server connection
    :param list datasource_names: a list of datasource names to filter the results by
    :return: pd.DataFrame
    """
    datasources_df = pd.DataFrame(get_all_datasource_fields(conn))
    if datasource_names:
        datasources_df = datasources_df[datasources_df['name'].isin(datasource_names)]
    return datasources_df


def get_datasource_connections_dataframe(conn, datasource_id) -> pd.DataFrame:
    datasource_connections_json = conn.query_data_source_connections(datasource_id).json()['connections']['connection']
    datasource_connections_df = pd.DataFrame(datasource_connections_json)
    return datasource_connections_df



import pandas as pd
import os


from tableau_api_lib.utils.querying import get_workbooks_dataframe, get_projects_dataframe, get_datasources_dataframe, \
    get_workbook_connections_dataframe, get_datasource_connections_dataframe
from tableau_api_lib.utils.filemod import modify_tableau_zipfile, create_temp_dirs, delete_temp_files
from tableau_api_lib.utils import flatten_dict_column, get_server_netloc
from tableau_api_lib.exceptions import ContentOverwriteDisabled


def get_source_datasource_df(conn_source):
    datasource_df = get_datasources_dataframe(conn_source)
    datasource_df = flatten_dict_column(datasource_df, keys=['id', 'name'], col_name='project')
    datasource_df = flatten_dict_column(datasource_df, keys=['id', 'name'], col_name='owner')
    datasource_df.columns = ["source_" + column for column in datasource_df.columns]
    datasource_df['source_project_name_lower'] = datasource_df['source_project_name'].apply(lambda x: x.lower())
    return datasource_df


def get_target_project_df(conn_target):
    project_df = get_projects_dataframe(conn_target)[['name', 'id']]
    project_df.columns = ['target_project_' + col for col in list(project_df.columns)]
    project_df['target_project_name_lower'] = project_df['target_project_name'].apply(lambda x: x.lower())
    return project_df


def get_mapped_datasource_df(source_datasource_df, target_project_df):
    mapped_datasource_df = source_datasource_df.merge(target_project_df,
                                                      how='left',
                                                      left_on='source_project_name_lower',
                                                      right_on='target_project_name_lower')
    relevant_cols = ['source_tags', 'source_description', 'source_encryptExtracts', 'source_id', 'source_isCertified',
                     'source_isPublished', 'source_name', 'source_type', 'source_useRemoteQueryAgent',
                     'source_serverName', 'target_project_name', 'target_project_id', 'source_owner_name']
    mapped_datasource_df = mapped_datasource_df[relevant_cols]
    return mapped_datasource_df


def get_all_workbook_connections_df(conn):
    all_wb_connections_df = pd.DataFrame()
    all_wbs = get_workbooks_dataframe(conn)
    for workbook_id in all_wbs['id']:
        wb_connections = get_workbook_connections_dataframe(conn, workbook_id).drop(columns=['id'])
        wb_connections = wb_connections[~wb_connections['userName'].isin([None, ''])]
        all_wb_connections_df = all_wb_connections_df.append(wb_connections, sort=False)
    all_wb_connections_df = all_wb_connections_df[['serverAddress', 'userName', 'serverPort']].drop_duplicates()
    return all_wb_connections_df


def get_all_datasource_connections_df(conn):
    all_ds_connections_df = pd.DataFrame()
    all_dss = get_datasources_dataframe(conn)
    for index, datasource in all_dss.iterrows():
        datasource_name, datasource_id = datasource['name'], datasource['id']
        ds_connections = get_datasource_connections_dataframe(conn, datasource_id).drop(columns=['id'])
        # ds_connections['datasource_name'] = datasource_name
        ds_connections = ds_connections[~ds_connections['userName'].isin([None, ''])]
        all_ds_connections_df = all_ds_connections_df.append(ds_connections, sort=False)
    all_ds_connections_df = all_ds_connections_df[['serverAddress', 'userName']].drop_duplicates()
    return all_ds_connections_df


def get_mapped_connections_df(conn_source, conn_target, source_connections_df):
    mapped_connections_df = source_connections_df.copy()
    for index, connection in mapped_connections_df.iterrows():
        if connection['serverAddress'] == get_server_netloc(conn_source.server):
            mapped_connections_df.at[index, 'target_serverAddress'] = get_server_netloc(conn_target.server)
        else:
            mapped_connections_df.at[index, 'target_serverAddress'] = connection['serverAddress']
    return mapped_connections_df


def get_datasource_connection_ports(all_datasource_connections_df, all_workbook_connections_df):
    all_datasource_connections_df = all_datasource_connections_df.merge(all_workbook_connections_df,
                                                                        how='left',
                                                                        left_on=['serverAddress', 'userName'],
                                                                        right_on=['serverAddress', 'userName'])
    return all_datasource_connections_df


def get_credentials_df(file_path):
    if os.path.exists(file_path):
        datasource_credentials_df = pd.read_csv(file_path)
        return datasource_credentials_df
    else:
        raise FileNotFoundError(f"No credentials CSV mapping file was found at '{file_path}'.")


def download_datasources(conn_source, datasource_df, download_dir=None):
    download_dir = download_dir or os.getcwd() + '/temp/datasources/source'
    for index, datasource in datasource_df.iterrows():
        with open(f"{download_dir}/{datasource['source_name']}.tdsx", 'wb') as file:
            print("downloading datasource '{0}' to '{1}/{0}.tdsx'...".format(datasource['source_name'], download_dir))
            response = conn_source.download_data_source(datasource_id=datasource['source_id'])
            file.write(response.content)


def get_credential_mappings(datasource_df, credentials_df):
    mapped_df = datasource_df.merge(credentials_df,
                                    how='left',
                                    left_on=['target_serverAddress', 'userName', 'serverPort'],
                                    right_on=['serverAddress', 'userName', 'serverPort'],
                                    suffixes=[None, '_delete'])
    cols_to_keep = list(datasource_df.columns) + ['password']
    return mapped_df[cols_to_keep]


def set_temp_dirs(temp_dir):
    temp_dir = temp_dir or os.getcwd() + '/temp'
    extraction_dir = temp_dir + '/extracted'
    create_temp_dirs(temp_dir)
    return temp_dir, extraction_dir


def get_cloning_df(conn_source, conn_target, credentials_file_path):
    source_datasource_df = get_source_datasource_df(conn_source)
    target_project_df = get_target_project_df(conn_target)
    mapped_datasource_df = get_mapped_datasource_df(source_datasource_df, target_project_df)
    source_connections_df = get_datasource_connection_ports(get_all_workbook_connections_df(conn_source),
                                                            get_all_datasource_connections_df(conn_source))
    mapped_connections_df = get_mapped_connections_df(conn_source, conn_target, source_connections_df)
    mapped_datasource_df = mapped_datasource_df.merge(mapped_connections_df, how='left',
                                                      left_on='source_serverName',
                                                      right_on='serverAddress')
    mapped_credentials_df = get_credential_mappings(mapped_datasource_df,
                                                    get_credentials_df(credentials_file_path))
    return mapped_credentials_df


def modify_datasources_by_project(conn_source, conn_target, source_project_dir, target_project_dir, extraction_dir):
    for file in os.listdir(source_project_dir):
        modify_tableau_zipfile(zipfile_path=source_project_dir + '/' + file,
                               conn_source=conn_source, conn_target=conn_target,
                               extraction_dir_path=extraction_dir,
                               destination_dir_path=target_project_dir)
        delete_temp_files(extraction_dir)


def publish_datasources_by_project(conn_target, project_datasources_df):
    for index, datasource in project_datasources_df.iterrows():
        response = conn_target.publish_data_source(datasource_file_path=datasource['file_path'],
                                                   datasource_name=datasource['source_name'],
                                                   project_id=datasource['target_project_id'],
                                                   connection_username=datasource['userName'],
                                                   connection_password=datasource['password'],
                                                   embed_credentials_flag=True if datasource['userName'] else False)
        if response.status_code in [200, 201]:
            project_datasources_df.at[index, 'target_id'] = response.json()['datasource']['id']
    return project_datasources_df


def update_datasources_by_project(conn_target, project_datasources_df):
    for index, datasource in project_datasources_df.iterrows():
        conn_target.update_data_source(datasource_id=datasource['target_id'],
                                       # new_owner_id=,
                                       is_certified_flag=datasource['source_isCertified'])
        if 'tag' in datasource['source_tags'].keys():
            tags = datasource['source_tags']['tag']
            tag_list = [tag_dict['label'] for tag_dict in tags]
            conn_target.add_tags_to_data_source(datasource_id=datasource['target_id'], tags=tag_list)


def clone_datasources_by_project(conn_source, conn_target, mapped_credentials_df, temp_dir, project, extraction_dir):
    source_project_dir = f"{temp_dir}/datasources/source/{project}"
    target_project_dir = f"{temp_dir}/datasources/target/{project}"
    os.makedirs(source_project_dir, exist_ok=False)
    os.makedirs(target_project_dir, exist_ok=False)
    project_datasources_df = mapped_credentials_df[mapped_credentials_df['target_project_name'] == project].copy()
    project_datasources_df['file_path'] = project_datasources_df['source_name'].apply(
        lambda file: f"{target_project_dir}/{file}.tdsx")
    project_datasources_df.fillna('', inplace=True)
    download_datasources(conn_source, project_datasources_df, download_dir=source_project_dir)
    modify_datasources_by_project(conn_source, conn_target, source_project_dir, target_project_dir, extraction_dir)
    project_datasources = publish_datasources_by_project(conn_target, project_datasources_df)
    update_datasources_by_project(conn_target, project_datasources)


def clone_datasources(conn_source, conn_target, credentials_file_path=None, temp_dir=None, overwrite_policy=None):
    temp_dir, extraction_dir = set_temp_dirs(temp_dir)
    mapped_credentials_df = get_cloning_df(conn_source, conn_target, credentials_file_path)
    for project in mapped_credentials_df['target_project_name'].unique():
        clone_datasources_by_project(conn_source, conn_target, mapped_credentials_df, temp_dir, project, extraction_dir)
    delete_temp_files(temp_dir)



from tableau_api_lib.api_requests import BaseRequest


class AddDatasourceToScheduleRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding datasources to schedules.
    :param class ts_connection: the Tableau Server connection object
    :param str datasource_id: the datasource ID
    """
    def __init__(self,
                 ts_connection,
                 datasource_id):
        super().__init__(ts_connection)
        self._datasource_id = datasource_id
        self.base_add_datasource_request()

    @property
    def required_datasource_param_keys(self):
        return ['id']

    @property
    def required_datasource_param_values(self):
        return [self._datasource_id]

    def base_add_datasource_request(self):
        self._request_body.update({'task': {'extractRefresh': {}}})
        return self._request_body

    def modified_add_datasource_request(self):
        self._request_body['task']['extractRefresh'].update({'datasource': {}})
        self._request_body['task']['extractRefresh']['datasource'].update(
            self._get_parameters_dict(self.required_datasource_param_keys,
                                      self.required_datasource_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_add_datasource_request()



from tableau_api_lib.api_requests import BaseRequest


class UpdateWorkbookNowRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating workbook extracts immediately.
    :param class ts_connection: the Tableau Server connection object
    """
    def __init__(self,
                 ts_connection):
        super().__init__(ts_connection)

    def base_update_workbook_now_request(self):
        return self._request_body

    def get_request(self):
        return self.base_update_workbook_now_request()



class InvalidRestApiVersion(Exception):
    """
    Exception for flagging API method calls to endpoints that do not exist for the (older) version being used.
    """
    def __init__(self, func, api_version_used, api_version_required):
        """
        Generates an error message citing the api_version_used and api_version_required.
        :param str api_version_used: the REST API being used by the active connection
        :param str api_version_required: the minimum REST API version required
        """
        error_message = """
        The REST API endpoint referenced in function '{0}' requires a minimum API version of {1}.
        The API version you are using is {2}, which pre-dates the endpoint referenced.
        Please visit Tableau's REST API reference to identify the methods relevant for your legacy API version.
        """.format(func.__name__, api_version_required, api_version_used)
        super().__init__(error_message)


class InvalidTableauServerQuery(Exception):
    """
    Exception for flagging invalid Tableau Server site queries.
    """
    def __init__(self, query_object, object_property):
        """
        Generates an error message for the query object and object property provided.
        :param str query_object: the object being queries ['site', 'project', 'user', etc.]
        :param str object_property: the property being queried ['id', 'name', etc.]
        """
        error_message = """
        The active Tableau Server {} query did not contain a property for "{}".
        Please check that your Tableau Server connection is valid and signed in.
        """.format(query_object, object_property)
        super().__init__(error_message)


class ContentNotFound(Exception):
    """
    Exception for situations where Tableau Server content is queried but does not exist on the site.
    """
    def __init__(self, content_type=None, content_id=None):
        if content_type:
            error_message = """
            The {0} (id='{1}') does not exist on the active site.
            """.format(content_type, content_id)
        else:
            error_message = """
            The queried Tableau Server content [users, projects, etc.] does not exist on the active site.
            """
        super().__init__(error_message)


class ContentOverwriteDisabled(Exception):
    """
    Exception for flagging that the target server has objects with the same name as those being cloned from the source.
    """
    def __init__(self, content_object):
        error_message = """
        The target connection has at least one named {} that already exists, and no 'overwrite_policy' was specified.
        If you'd like to overwrite the existing content, set the 'overwrite_policy' value to 'overwrite'.
        """.format(content_object)
        super().__init__(error_message)


class UsersNotFound(Exception):
    """
    Exception for situations where Tableau Server users are queried but non exist on the site.
    """
    def __init__(self):
        error_message = """
        An attempt was made to query Tableau Server users, but none exist for the active site.
        Make sure users have been created on the site and try again.
        """
        super().__init__(error_message)


class PaginationError(Exception):
    """
    Exception for when users attempt extract pages from a non-paginated REST API response.
    """
    def __init__(self, func):
        error_message = """
        The Tableau Server REST API method {} does not return paginated results. 
        The extract_pages() method wrapping this call is intended for paginated results only.
        Not all Tableau Server REST API methods support pagination. 
        Try calling on this method directly rather than using extract_pages().
        """.format(func.__name__)
        super().__init__(error_message)



from tableau_api_lib.api_requests import BaseRequest


class AddUserToAlertRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding users to data alerts.
    :param class ts_connection: the Tableau Server connection object
    :param str user_id: the user ID for the user being added to the alert
    """
    def __init__(self,
                 ts_connection,
                 user_id):
        super().__init__(ts_connection)
        self._user_id = user_id

    def base_add_user_request(self):
        self._request_body.update({'user': {'id': self._user_id}})
        return self._request_body

    def get_request(self):
        return self.base_add_user_request()



"""
workflow logic:
1) get workbook details for the source site
2) get workbook details for the target site
3) get overlapping workbooks (source & target)
4) if overwrite_policy == 'overwrite', then delete the overlapping workbooks on the target site
5) if overwrite_policy == None, then throw an exception
6) if overwrite_policy == 'merge', then proceed with cloning the non-overlapping workbooks and leave the others as is
7) for each workbook on the source site, build a clone of that workbook on the target site
    7a) for each workbook, query its connections
    7b) for each connection, check if it is a published data source
    7c) build a dataframe that labels datasources as embedded or published
    7d) for each published data source on the source site, download and publish that data source to the target site
            must open each file and edit its server, username, and site name
    7e) make sure each data source is published to the correct respective workbook
    7f) download all workbooks (will be downloaded as .twbx)
        save each workbook to the local file system (in the future provide a way to save to cloud storage)
    7g) publish all 'published' datasources to the target site
    7h) publish all workbooks to their respective workbooks, with published data source credentials
            for published data sources: server name, connection username, connection password must be provided

"""


import pandas as pd
import requests

from tableau_api_lib.utils.querying import get_workbooks_dataframe, get_projects_dataframe, get_datasources_dataframe
from tableau_api_lib.utils import flatten_dict_column
from tableau_api_lib.exceptions import ContentOverwriteDisabled


def get_source_workbook_df(conn_source, workbook_names=None):
    """
    Query details for all workbooks on the source site, or only the workbooks listed in 'workbook_names'.
    :param class conn_source: the Tableau Server connection object
    :param list workbook_names: a list of the desired workbook names whose details will be queried
    :return: Pandas DataFrame
    """
    workbook_df = get_workbooks_dataframe(conn_source)
    workbook_df.fillna(value='', inplace=True)
    if workbook_names:
        workbook_df = workbook_df[workbook_df['name'].isin(workbook_names)]
    workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name'], col_name='project')
    workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name', 'email'], col_name='owner')
    workbook_df.columns = ["source_" + column for column in workbook_df.columns]
    return workbook_df


def get_target_workbook_df(conn_target, workbook_names=None):
    """
    Query details for all workbooks on the source site, or only the workbooks listed in 'workbook_names'.
    :param class conn_target: the Tableau Server connection object
    :param list workbook_names: a list of the desired workbook names whose details will be queried
    :return: Pandas DataFrame
    """
    try:
        workbook_df = get_workbooks_dataframe(conn_target)
        workbook_df.fillna(value='', inplace=True)
        if workbook_names:
            workbook_df = workbook_df[workbook_df['name'].isin(workbook_names)]
        workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name'], col_name='project')
        workbook_df = flatten_dict_column(workbook_df, keys=['id', 'name', 'email'], col_name='owner')
        workbook_df.columns = ["target_" + column for column in workbook_df.columns]
    except ValueError:
        workbook_df = pd.DataFrame()
    return workbook_df


def get_source_to_target_wb_df(source_workbook_df, conn_target):
    """
    Get a DataFrame with source workbook details, source user details, and source target details combined.
    :param pd.DataFrame source_workbook_df: the source workbook details
    :param class conn_target: the target Tableau Server connection
    :return DataFrame: workbook_details_df
    """
    target_projects_df = get_projects_dataframe(conn_target)
    target_projects_df.columns = ["target_project_" + column for column in target_projects_df.columns]
    workbook_details_df = source_workbook_df.merge(target_projects_df,
                                                   how='left',
                                                   left_on='source_project_name',
                                                   right_on='target_project_name')
    return workbook_details_df


def get_source_to_target_ds_df(conn_target, source_datasource_df):
    """
    Get a DataFrame with source datasource details, source user details, and source target details combined.
    :param pd.DataFrame source_datasource_df: the source datasource details
    :param class conn_target: the target Tableau Server connection
    :return DataFrame: datasource_details_df
    """
    target_projects_df = get_projects_dataframe(conn_target)
    target_projects_df.columns = ["target_project_" + column for column in target_projects_df.columns]
    datasource_details_df = source_datasource_df.merge(target_projects_df,
                                                       how='left',
                                                       left_on='project_name',
                                                       right_on='target_project_name')
    return datasource_details_df


def add_connection_context(workbook, workbook_connections_df):
    workbook_connections_df['workbook_name'] = workbook['source_name']
    workbook_connections_df['workbook_id'] = workbook['source_id']
    workbook_connections_df['project_name'] = workbook['source_project_name']
    workbook_connections_df['project_id'] = workbook['source_project_id']
    return workbook_connections_df


def get_workbook_names(workbook_df, column_name) -> list:
    """
    Get the workbook names for the active site on the specified Tableau Server connection.
    :param DataFrame workbook_df: the workbook details DataFrame for the server
    :param str column_name: the name of the column containing workbook names [source_workbook_name, target_workbook_name]
    :return: workbook names
    """
    workbook_names = list(workbook_df[column_name])
    workbook_names = [workbook_name for workbook_name in workbook_names]
    return workbook_names


def get_overlapping_workbook_names(source_workbook_df, target_workbook_df) -> list:
    try:
        source_workbook_names = set(get_workbook_names(source_workbook_df, 'source_workbook_name'))
        target_workbook_names = set(get_workbook_names(target_workbook_df, 'target_workbook_name'))
        overlapping_workbook_names = source_workbook_names.intersection(target_workbook_names)
        return list(overlapping_workbook_names)
    except KeyError:
        return []


def delete_workbooks(conn, workbook_details_df, workbook_names):
    print("deleting overlapping target workbooks...")
    responses = []
    workbooks_to_delete = workbook_details_df[workbook_details_df['target_workbook_name'].isin(workbook_names)]
    for i, workbook in workbooks_to_delete.iterrows():
        print("deleting workbook: (name='{}', id='{}'".
              format(workbook['target_workbook_id'], workbook['target_workbook_name']))
        responses.append(conn.delete_workbook(workbook_id=workbook['target_workbook_id']))
    print("overlapping target workbooks deleted")
    return conn


def validate_inputs(overwrite_policy):
    valid_overwrite_policies = [
        None,
        'overwrite',
        'merge'  # add logic for merging (leave existing content as is and create any missing content)
    ]
    if overwrite_policy in valid_overwrite_policies:
        pass
    else:
        raise ValueError("Invalid overwrite policy provided: '{}'".format(overwrite_policy))


# def get_published_workbook_datasources(conn_source, source_workbook_df):
#     published_workbook_datasources_df = pd.DataFrame()
#     for index, workbook in source_workbook_df.iterrows():
#         workbook_connections_df = pd.DataFrame(
#             conn_source.query_workbook_connections(workbook['source_id']).json()['connections']['connection'])
#         workbook_connections_df = extract_datasource_details(workbook_connections_df, ['id', 'name'])
#         workbook_connections_df = add_connection_context(workbook, workbook_connections_df)
#         published_workbook_datasources_df = pd.concat([published_workbook_datasources_df, workbook_connections_df],
#                                                       sort=False)
#     published_workbook_datasources_df = published_workbook_datasources_df[
#         published_workbook_datasources_df['serverAddress'].isin([urlparse(conn_source.server).netloc])]
#     all_datasources_df = get_datasources_dataframe(conn_source)
#     print("published wb ds: \n: ", published_workbook_datasources_df['datasource_id'])
#     print("all ds: \n: ", all_datasources_df['id'])
#     published_workbook_datasources_df = published_workbook_datasources_df[
#         published_workbook_datasources_df['datasource_id'].isin(all_datasources_df['id'])]
#     return published_workbook_datasources_df


def get_published_datasources(conn_source):
    published_datasources_df = get_datasources_dataframe(conn_source)
    published_datasources_df = flatten_dict_column(published_datasources_df, keys=['id', 'name'], col_name='project')
    published_datasources_df = flatten_dict_column(published_datasources_df, keys=['id', 'name'], col_name='owner')
    return published_datasources_df


def download_published_datasources(conn_source, published_datasources_df, file_path='temp/'):
    projects = published_datasources_df['project_name'].unique()
    published_datasources_df['datasource_file_path'] = published_datasources_df['name']\
        .apply(lambda x: file_path + str(x) + '.tdsx')
    print("published_datasources_df_cols: \n", published_datasources_df.columns)
    for project in projects:
        print("downloading datasources from project '{}'...".format(project))
        datasources_df = published_datasources_df[published_datasources_df['project_name'].isin([project])]
        for index, datasource in datasources_df.iterrows():
            print("downloading datasource '{}'...".format(datasource['name']))
            file_contents = conn_source.download_data_source(datasource_id=datasource['id'])
            with open(datasource['datasource_file_path'], 'wb') as file:
                file.write(file_contents.content)
            print("completed download of datasource '{}'".format(datasource['name']))
        print("completed downloading datasources from project '{}'".format(project))
    return published_datasources_df


def publish_datasources(conn_target, datasources_df):
    publish_responses = []
    datasources_df['target_datasource_id'] = ''
    for index, datasource in datasources_df.iterrows():
        print("publishing datasource '{}'...".format(datasource['name']))
        response = conn_target.publish_data_source(datasource_file_path=datasource['datasource_file_path'],
                                                   datasource_name=datasource['name'],
                                                   project_id=datasource['target_project_id'])
        if response.status_code == 201:
            datasources_df.loc[index, 'target_datasource_id'] = response.json()['datasource']['id']
            print("successfully published datasource '{}'".format(datasource['name']))
        publish_responses.append(response)
    return datasources_df, publish_responses


def clone_datasource_tags(conn_target, datasources_df):
    responses = []
    for index, datasource in datasources_df.iterrows():
        print("cloning tags for datasource '{}'...".format(datasource['name']))
        try:
            datasource_tags = [tag['label'] for tag in datasource['tags']['tag']]
            response = conn_target.add_tags_to_data_source(datasource['target_datasource_id'], datasource_tags)
            responses.append(response)
            if response.status_code == 200:
                print("successfully cloned tags for datasource '{}'".format(datasource['name']))
            else:
                print("failed to clone tags for datasource '{}'".format(datasource['name']))
        except KeyError:
            print("no tags exist for datasource '{}'".format(datasource['name']))
    return responses


def process_overlapping_workbooks(conn_target,
                                  target_workbook_df,
                                  overlapping_workbook_names,
                                  overwrite_policy) -> None:
    """
    Processes overlapping workbooks depending on the overwrite policy specified.
    :param class conn_target: the target Tableau Server connection
    :param pd.DataFrame target_workbook_df: a Pandas DataFrame with target workbook details
    :param list overlapping_workbook_names: a list of workbooks present on both the source and target connections
    :param str overwrite_policy: must be set to 'overwrite' to enable overwriting existing content
    :return:
    """
    if any(overlapping_workbook_names) and not overwrite_policy:
        raise ContentOverwriteDisabled('workbooks')
    if any(overlapping_workbook_names) and overwrite_policy == 'overwrite':
        delete_workbooks(conn_target, target_workbook_df, overlapping_workbook_names)
    pass


# TODO: handle downloading and publishing for .tdsx and .twbx / .twb files
def clone_workbooks(conn_source, conn_target, workbook_names=None, overwrite_policy=None):
    """
    Clones workbooks from the source server to the target server.
    :param class conn_source: the source server connection
    :param class conn_target: the target server connection
    :param list workbook_names: (optional) a list of workbook names to clone; specifying no names clones all workbooks
    :param str overwrite_policy: (optional) set to 'overwrite' to overwrite content; defaults to not overwriting
    :return: None
    """
    validate_inputs(overwrite_policy)
    source_workbook_df = get_source_workbook_df(conn_source=conn_source, workbook_names=workbook_names)
    target_workbook_df = get_target_workbook_df(conn_target=conn_target, workbook_names=workbook_names)
    # refactor the overlapping checks to a new function: process_overwrite_policy()
    overlapping_workbook_names = get_overlapping_workbook_names(source_workbook_df, target_workbook_df)
    process_overlapping_workbooks(conn_target, target_workbook_df, overlapping_workbook_names, overwrite_policy)
    workbook_details_df = get_source_to_target_wb_df(source_workbook_df, conn_target)
    workbook_details_df.to_csv('delete_workbook_details_df.csv')
    # published_datasources_df = get_published_datasources(conn_source)
    # published_datasources_df = download_published_datasources(conn_source, published_datasources_df)
    # -> modify the downloaded data source .tds files to reference the target server instead of the source server
    # -> save the modified .tds files and zip them back into .tdsx files
    # published_datasources_df = get_source_to_target_ds_df(conn_target, published_datasources_df)
    # published_datasources_df, publish_responses = publish_datasources(conn_target, published_datasources_df)
    # clone_tag_responses = clone_datasource_tags(conn_target, published_datasources_df)
    # -> download source site's workbooks
    # -> modify downloaded workbook .twb files to reflect the target server instead of the source server
    # -> repackage files if they were .twbx, otherwise leave as .twb
    # -> publish source site's workbooks to target site
    # -> update workbooks to have the correct owner (and other metadata)
    #   -> optional mapping stage: map_workbook_owners (searches for original owner on target site, otherwise default)



from tableau_api_lib.api_endpoints import BaseEndpoint


class DataAlertEndpoint(BaseEndpoint):
    def __init__(self, 
                 ts_connection,
                 query_data_alerts=False,
                 query_data_alert=False,
                 data_alert_id=None, 
                 user_id=None, 
                 add_user=False, 
                 remove_user=False,
                 parameter_dict=None):
        """
        Data Alert endpoint for Tableau Server API api_requests.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_data_alerts: True if querying all data alerts, False otherwise
        :param bool query_data_alert: True if querying a specific data alert, False otherwise
        :param str data_alert_id: the data alert ID
        :param str user_id: the user ID
        :param bool add_user: True if adding a user to the alert, False otherwise
        :param bool remove_user: True if removing a user from the alert; False otherwise
        :param dict parameter_dict: dictionary of URL parameters to append; the value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """
        
        super().__init__(ts_connection)
        self._query_data_alerts = query_data_alerts
        self._query_data_alert = query_data_alert
        self._data_alert_id = data_alert_id
        self._user_id = user_id
        self._add_user = add_user
        self._remove_user = remove_user
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_data_alerts,
            self._query_data_alert,
            self._add_user,
            self._remove_user
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_data_alert_url(self):
        return "{0}/api/{1}/sites/{2}/dataAlerts".format(self._connection.server,
                                                         self._connection.api_version,
                                                         self._connection.site_id)
    
    @property
    def base_data_alert_id_url(self):
        return "{0}/{1}".format(self.base_data_alert_url, 
                                self._data_alert_id)
    
    @property
    def base_data_alert_user_url(self):
        return "{0}/users".format(self.base_data_alert_id_url)
    
    @property
    def base_data_alert_user_id_url(self):
        return "{0}/{1}".format(self.base_data_alert_user_url, 
                                self._user_id)
    
    def get_endpoint(self):
        url = None
        if not (self._data_alert_id or self._user_id or self._add_user or self._remove_user):
            url = self.base_data_alert_url
        elif self._data_alert_id and not (self._user_id or self._add_user or self._remove_user):
            url = self.base_data_alert_id_url
        elif self._data_alert_id and self._add_user and self._user_id and not self._remove_user:
            url = self.base_data_alert_user_url
        elif self._data_alert_id and self._remove_user and self._user_id and not self._add_user:
            url = self.base_data_alert_user_id_url
        else:
            self._invalid_parameter_exception()
            
        return self._append_url_parameters(url)



sample_config = {
    'tableau_prod': {
        'server': 'https://<YOUR_SERVER>.com',
        'api_version': '<YOUR_API_VERSION>',
        'username': '<YOUR_USERNAME>',
        'password': '<YOUR_PASSWORD>',
        # comment out username / password and uncomment the lines below to use personal access token authentication
        # 'personal_access_token_name': '<YOUR_PERSONAL_ACCESS_TOKEN_NAME>',
        # 'personal_access_token_secret': '<YOUR_PERSONAL_ACCESS_TOKEN_SECRET>',
        'site_name': '<YOUR_SITE_NAME>',
        'site_url': '<YOUR_SITE_URL>'
    }
}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tableau_api_lib",
    version="0.0.74",
    author="Elliott Stam",
    author_email="elliott.stam@gmail.com",
    description="This library enables developers to call any method seen in Tableau Server's REST API documentation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/divinorum-webb/tableau-api-lib",
    packages=setuptools.find_packages('src'),
    package_dir={
        '': 'src',
    },
    install_requires=[
        'requests>2',
        'urllib3',
        'pandas',
        'requests-toolbelt>=0.9.0',
        'bleach>=3.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




from .custom_exceptions import InvalidParameterException, InvalidFileTypeException
from .tableau_server_exceptions import InvalidRestApiVersion, InvalidTableauServerQuery, ContentOverwriteDisabled, \
    ContentNotFound, PaginationError, UsersNotFound



import requests
import warnings
from functools import wraps
from tableau_api_lib.exceptions import InvalidRestApiVersion


def verify_response(success_code):
    """
    Verifies that the request being executed has a response whose status code matches the success code provided.
    This decorates the function calls which send API api_requests to the server.

    :param success_code:    The code that will be passed with the server's response body if the request was successful.
    :type success_code:     int
    :return:                Returns the results of the function this decorates.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            response = requests.get(self.request_url)
            if response.status_code != success_code:
                raise Exception('The request to Tableau Server returned code \n'
                                ' {} instead of {} in function {}'.format(response.status_code,
                                                                          success_code,
                                                                          func.__name__))
            return func(self, *args, **kwargs)
        return wrapper
    return decorator


def verify_signed_in(func):
    """
    Verifies that the Tableau Server object is signed in.
    This decorates the function calls which send API api_requests to the server.

    :param func:    The code that will be passed with the server's response body if the request was successful.
    :type func:     function
    :return:        Returns the results of the function this decorates.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.auth_token:
            return func(self, *args, **kwargs)
        else:
            raise Exception('The Tableau Server connection is not logged in.')
    return wrapper


def verify_config_variables(func):
    """
    Verifies that the required config variables exist.
    This decorates the function calls which send API api_requests to the server.

    :param func:    The code that will be passed with the server's response body if the request was successful.
    :type func:     function
    :return:        Returns the results of the function this decorates.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        required_config_variables = [
            'server',
            'api_version',
            'site_name',
            'site_url'
        ]
        if (self._config and self._env) and type(self._config) == dict:
            config_variables = self._config[self._env].keys()
            missing_required_variables = [variable for variable in required_config_variables
                                          if variable not in config_variables]
            if not any(missing_required_variables):
                return func(self, *args, **kwargs)
            else:
                raise ValueError("""
                The configuration variables provided are invalid.
                Please provide the following missing configuration variables: {}
                """.format(missing_required_variables))
        raise Exception('Please provide a configuration dict to establish a connection.')
    return wrapper


def verify_rest_api_version(func):
    """
    Verifies that the Tableau Server object is signed in.
    This decorates the function calls which send API api_requests to the server.

    :param func:    The code that will be passed with the server's response body if the request was successful.
    :type func:     function
    :return:        Returns the results of the function this decorates.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if (self._config and self._env) and type(self._config) == dict:
            try:
                server_info = self.server_info().json()
                server_api_version = server_info['serverInfo']['restApiVersion']
                if self._config[self._env]['api_version'] > server_api_version:
                    raise Exception("""Your API version is too damn high!""")
                if self._config[self._env]['api_version'] < server_api_version:
                    warnings.warn("""
                    WARNING:
                    The Tableau Server REST API version you specified is lower than the version your server uses.
                    Your Tableau Server is on REST API version {0}.
                    The REST API version you specified is {1}.
                    For optimal results, please change the 'api_version' config variable to {0}.
                    """.format(server_api_version, self._config[self._env]['api_version']))
            except AttributeError:
                warnings.warn("""
                Warning: could not verify your Tableau Server's API version.
                If using a legacy version of Tableau Server, be sure to reference the legacy Tableau Server
                REST API documentation provided by Tableau.
                Some current API methods may exist that are not available on your legacy Tableau Server.
                """)
            return func(self, *args, **kwargs)
        else:
            raise Exception('The Tableau Server connection is not logged in.')
    return wrapper


def verify_api_method_exists(version_introduced):
    """
    Verifies that the connection's REST API version is equal to or greater than the version the method was introduced.

    :param str version_introduced: the REST API version when the method was introduced
    :return: decorated function
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            conn_api_version = self._config[self._env]['api_version']
            if conn_api_version < version_introduced:
                raise InvalidRestApiVersion(func,
                                            api_version_used=conn_api_version,
                                            api_version_required=version_introduced)
            return func(self, *args, **kwargs)
        return wrapper
    return decorator



from .base_endpoint import BaseEndpoint
from .auth_endpoint import AuthEndpoint
from .data_alert_endpoint import DataAlertEndpoint
from .datasource_endpoint import DatasourceEndpoint
from .favorites_endpoint import FavoritesEndpoint
from .file_upload_endpoint import FileUploadEndpoint
from .flow_endpoint import FlowEndpoint
from .group_endpoint import GroupEndpoint
from .jobs_endpoint import JobsEndpoint
from .permissions_endpoint import PermissionsEndpoint
from .project_endpoint import ProjectEndpoint
from .schedules_endpoint import SchedulesEndpoint
from .site_endpoint import SiteEndpoint
from .subscriptions_endpoint import SubscriptionsEndpoint
from .tasks_endpoint import TasksEndpoint
from .user_endpoint import UserEndpoint
from .view_endpoint import ViewEndpoint
from .workbook_endpoint import WorkbookEndpoint
from .database_endpoint import DatabaseEndpoint
from .table_endpoint import TableEndpoint
from .column_endpoint import ColumnEndpoint
from .dqwarning_endpoint import DQWarningEndpoint
from .encryption_endpoint import EncryptionEndpoint
from .graphql_endpoint import GraphqlEndpoint



from tableau_api_lib.api_requests import BaseRequest


class UpdateWorkbookRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating workbooks.
    :param class ts_connection: the Tableau Server connection object
    :param bool show_tabs_flag: (optional) true if the workbook will show views as tabs, False otherwise
    :param str project_id: (optional) the ID of a project to assign the workbook to
    :param str owner_id: (optional) the ID of the user who will own the workbook
    """
    def __init__(self,
                 ts_connection,
                 show_tabs_flag=None,
                 project_id=None,
                 owner_id=None):

        super().__init__(ts_connection)
        self._show_tabs_flag = show_tabs_flag
        self._project_id = project_id
        self._owner_id = owner_id
        self.base_update_workbook_request()

    @property
    def optional_workbook_param_keys(self):
        return ['showTabs']

    @property
    def optional_project_param_keys(self):
        return ['id']

    @property
    def optional_owner_param_keys(self):
        return ['id']

    @property
    def optional_workbook_param_values(self):
        return ['true' if self._show_tabs_flag is True else 'false' if self._show_tabs_flag is False else None]

    @property
    def optional_project_param_values(self):
        return [self._project_id]

    @property
    def optional_owner_param_values(self):
        return [self._owner_id]

    def base_update_workbook_request(self):
        self._request_body.update({'workbook': {}})
        return self._request_body

    def modified_update_workbook_request(self):
        if any(self.optional_workbook_param_values):
            self._request_body['workbook'].update(
                self._get_parameters_dict(self.optional_workbook_param_keys,
                                          self.optional_workbook_param_values))

        if any(self.optional_project_param_values):
            self._request_body['workbook'].update({'project': {}})
            self._request_body['workbook']['project'].update(
                self._get_parameters_dict(self.optional_project_param_keys,
                                          self.optional_project_param_values))

        if any(self.optional_owner_param_values):
            self._request_body['workbook'].update({'owner': {}})
            self._request_body['workbook']['owner'].update(
                self._get_parameters_dict(self.optional_owner_param_keys,
                                          self.optional_owner_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_update_workbook_request()



"""
Logical Flow:
1) get groups from source connection
2) get groups from target connection
3) get overlapping groups
4) if overwrite policy = 'overwrite', then delete the overlapping groups from the target connection
5) if overwrite policy = None, then raise an exception
6) create groups on the target connection to mirror those on the source connection
7) * for future: figure out how to best populate the groups... configurable for Active Directory, by users, etc.
"""


import pandas as pd

from tableau_api_lib.exceptions import ContentOverwriteDisabled
from tableau_api_lib.utils.querying import get_groups_dataframe, get_group_users_dataframe
from tableau_api_lib.utils.querying.users import get_users_dataframe


def get_source_group_df(conn_source, group_names=None) -> pd.DataFrame:
    group_df = get_groups_dataframe(conn_source)
    if group_names:
        group_df = group_df[group_df['name'].isin(group_names)]
    group_df.rename(columns={
        'name': 'source_name',
        'id': 'source_id',
        'domain': 'source_domain',
        'userCount': 'source_user_count'
    }, inplace=True)
    return group_df


def get_target_group_df(conn_source, group_names=None) -> pd.DataFrame:
    group_df = get_groups_dataframe(conn_source)
    if group_names:
        group_df = group_df[group_df['name'].isin(group_names)]
    group_df.rename(columns={
        'name': 'target_name',
        'id': 'target_id',
        'domain': 'target_domain',
        'userCount': 'target_user_count'
    }, inplace=True)
    return group_df


def get_overlapping_group_names(source_group_df, target_group_df) -> list:
    source_group_names = set(source_group_df['source_name'])
    target_group_names = set(target_group_df['target_name'])
    overlapping_group_names = source_group_names.intersection(target_group_names)
    return list(overlapping_group_names)


def delete_groups(conn, target_group_df) -> list:
    print("deleting overlapping target groups...")
    group_ids = target_group_df['target_id']
    responses = [conn.delete_group(group_id) for group_id in group_ids]
    print("overlapping target groups deleted")
    return responses


def create_groups(conn_target, source_group_df) -> list:
    responses = []
    for i, _ in enumerate(source_group_df['source_name']):
        response = conn_target.create_group(new_group_name=source_group_df['source_name'][i])
        responses.append(response)
    return responses


def clone_group_users(conn_source, conn_target, source_group_df, target_group_df):
    combined_group_df = source_group_df.merge(target_group_df,
                                              how='left',
                                              left_on='source_name',
                                              right_on='target_name',
                                              suffixes=(None, None))
    target_users_df = get_users_dataframe(conn_target)
    for group_index, group in combined_group_df.iterrows():
        if group['source_name'] == 'All Users':
            continue
        source_group_users_df = get_group_users_dataframe(conn_source, group['source_id'])
        combined_group_users_df = source_group_users_df.merge(target_users_df,
                                                              how='left',
                                                              left_on='name',
                                                              right_on='name',
                                                              suffixes=(None, '_target'))
        for user_index, user in combined_group_users_df.iterrows():
            print("adding user to group: id={}  name={}".format(group['target_id'], group['target_name']))
            conn_target.add_user_to_group(group_id=group['target_id'], user_id=user['id_target'])


def process_overlapping_group_names(conn_target,
                                    target_group_df,
                                    overlapping_group_names,
                                    overwrite_policy) -> None:
    if any(overlapping_group_names) and not overwrite_policy:
        raise ContentOverwriteDisabled('groups')
    if any(overlapping_group_names) and overwrite_policy == 'overwrite':
        delete_groups(conn_target, target_group_df[target_group_df['target_name'].isin(overlapping_group_names)])
    pass


def reset_connection(conn):
    conn.sign_out()
    conn.sign_in()
    return conn


def clone_groups(conn_source,
                 conn_target,
                 group_names=None,
                 populate_users=False,
                 group_mapping_file=None,
                 overwrite_policy=None) -> list:
    source_group_df = get_source_group_df(conn_source, group_names)
    target_group_df = get_target_group_df(conn_target, group_names)
    overlapping_group_names = get_overlapping_group_names(source_group_df, target_group_df)
    process_overlapping_group_names(conn_target, target_group_df, overlapping_group_names, overwrite_policy)
    cloned_groups = create_groups(conn_target, source_group_df)
    if populate_users:
        reset_connection(conn_target)
        clone_group_users(conn_source, conn_target, source_group_df, get_target_group_df(conn_target, group_names))
    return cloned_groups



from tableau_api_lib.api_endpoints import BaseEndpoint


class TableEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_table=False,
                 query_tables=False,
                 update_table=False,
                 remove_table=False,
                 table_id=None):
        """
        Builds API endpoints for REST API table methods.
        :param class ts_connection: the Tableau Server connection object
        :param query_table: True if querying a specific database table, False otherwise
        :param query_tables: True if querying all database tables, False otherwise
        :param update_table: True if updating a specific database table, False otherwise
        :param remove_table: True if removing a specific database table, False otherwise
        :param table_id: the database table ID
        """

        super().__init__(ts_connection)
        self._query_table = query_table
        self._query_tables = query_tables
        self._update_table = update_table
        self._remove_table = remove_table
        self._table_id = table_id
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_table,
            self._query_tables,
            self._update_table,
            self._remove_table
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_table_url(self):
        return "{0}/api/{1}/sites/{2}/tables".format(self._connection.server,
                                                     self._connection.api_version,
                                                     self._connection.site_id)

    @property
    def base_table_id_url(self):
        return "{0}/{1}".format(self.base_table_url,
                                self._table_id)

    def get_endpoint(self):
        if self._table_id:
            url = self.base_table_id_url
        else:
            url = self.base_table_url
        if url:
            return url
        else:
            self._invalid_parameter_exception()



from tableau_api_lib.api_requests import BaseRequest


class UpdateSubscriptionRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating subscriptions.
    :param class ts_connection: the Tableau Server connection object
    :param str new_subscription_subject: (optional) a new subject for the subscription
    :param str new_schedule_id: (optional) the ID of a schedule to associate this subscription with
    :param bool attach_image_flag: True if attaching a .png image to the subscription, defaults to False
    :param bool attach_pdf_flag: True if attaching a .pdf file to the subscription, defaults to False
    """
    def __init__(self,
                 ts_connection,
                 new_subscription_subject=None,
                 new_schedule_id=None,
                 attach_image_flag=None,
                 attach_pdf_flag=None):

        super().__init__(ts_connection)
        self._new_subscription_subject = new_subscription_subject
        self._new_schedule_id = new_schedule_id
        self._attach_image_flag = attach_image_flag
        self._attach_pdf_flag = attach_pdf_flag

    @property
    def optional_param_keys(self):
        return [
            'attachImage',
            'attachPdf'
        ]

    @property
    def optional_param_values(self):
        return [
            self._attach_image_flag,
            self._attach_pdf_flag
        ]

    def base_update_subscription_request(self):
        if self._new_subscription_subject and self._new_schedule_id:
            self._request_body.update({
                'subscription': {
                    'subject': self._new_subscription_subject,
                    'schedule': {'id': self._new_schedule_id}
                }
            })
        elif self._new_subscription_subject and not self._new_schedule_id:
            self._request_body.update({
                'subscription': {
                    'subject': self._new_subscription_subject
                }
            })
        else:
            self._request_body.update({
                'subscription': {
                    'schedule': {'id': self._new_schedule_id}
                }
            })
        if any(self.optional_param_values):
            self._request_body['subscription'].update(self._get_parameters_dict(self.optional_param_keys,
                                                                                self.optional_param_values))
        return self._request_body

    def get_request(self):
        return self.base_update_subscription_request()



"""
The functions below enable you to clone the schedules from Server A to Server B.
This is particularly useful when doing site exports / imports.
The only functions are for use outside of this module:

clone_schedules
clone_schedule_state
override_schedule_state
"""
from tableau_api_lib.utils import extract_pages
from tableau_api_lib.exceptions import InvalidParameterException


def get_schedule_ids(conn, schedule_names=None):
    if schedule_names:
        schedule_ids = [schedule['id'] for schedule in extract_pages(conn.query_schedules)
                        if schedule['name'] in schedule_names]
    else:
        schedule_ids = [schedule['id'] for schedule in extract_pages(conn.query_schedules)]
    return schedule_ids


def get_schedule_details(conn, schedule_ids):
    schedule_details = [conn.update_schedule(schedule_id).json()['schedule'] for schedule_id in schedule_ids]
    return schedule_details


def create_base_schedules(destination_conn, schedules, clone_name_prefix=None):
    responses = []
    for schedule in schedules:
        response = destination_conn.create_schedule(schedule_name=clone_name_prefix + schedule['name'],
                                                    schedule_priority=schedule['priority'],
                                                    schedule_type=schedule['type'],
                                                    schedule_execution_order=schedule['executionOrder'],
                                                    schedule_frequency=schedule['frequency'],
                                                    start_time=schedule['frequencyDetails']['start'],
                                                    end_time=schedule['frequencyDetails'].get('end', None),
                                                    interval_expression_list=
                                                    schedule['frequencyDetails']['intervals']['interval'])
        responses.append(response.json()['schedule'])
    return responses


def update_schedule_state(destination_conn, original_schedules, destination_schedules):
    responses = []
    for i, schedule in enumerate(destination_schedules):
        response = destination_conn.update_schedule(schedule_id=destination_schedules[i]['id'],
                                                    schedule_state=original_schedules[i]['state'])
        responses.append(response.json()['schedule'])
    return responses


def is_valid_state(new_state):
    new_state = new_state.lower().capitalize() if new_state else None
    if new_state in ['Suspended', 'Active', None]:
        return True
    else:
        return False


def override_schedule_state(conn, new_state, schedule_names=None):

    new_state = new_state.lower().capitalize() if new_state else None
    if not is_valid_state(new_state):
        raise InvalidParameterException('override_schedule_state()', new_state)

    if new_state:
        responses = []
        schedule_ids = get_schedule_ids(conn, schedule_names)
        for schedule_id in schedule_ids:
            response = conn.update_schedule(schedule_id=schedule_id, schedule_state=new_state)
            responses.append(response.json()['schedule'])
        return responses
    else:
        pass


def clone_schedule_state(origin_conn,
                         destination_conn,
                         origin_schedule_names=None,
                         destination_schedule_names=None):

    if origin_schedule_names and destination_schedule_names:
        if len(origin_schedule_names) != len(destination_schedule_names):
            print("When specifying schedule names, \
            you must provide an equal number of names for both the origin and destination.")
            raise InvalidParameterException('clone_schedule_state()',
                                            (origin_schedule_names, destination_schedule_names))
    elif origin_schedule_names or destination_schedule_names:
        raise InvalidParameterException('clone_schedule_state()',
                                        (origin_schedule_names, destination_schedule_names))

    origin_schedule_ids = get_schedule_ids(origin_conn, schedule_names=origin_schedule_names)
    origin_schedule_details = get_schedule_details(origin_conn, schedule_ids=origin_schedule_ids)
    destination_schedule_ids = get_schedule_ids(destination_conn, schedule_names=destination_schedule_names)
    destination_schedule_details = get_schedule_details(destination_conn, schedule_ids=destination_schedule_ids)
    updated_schedules = update_schedule_state(destination_conn, origin_schedule_details, destination_schedule_details)
    print(updated_schedules)
    return updated_schedules


def clone_schedules(origin_conn,
                    destination_conn,
                    state_override=None,
                    schedule_names=None,
                    clone_name_prefix=None):

    state_override = state_override.lower().capitalize() if state_override else None

    if not is_valid_state(state_override):
        raise InvalidParameterException('clone_schedules()', state_override)

    origin_schedule_ids = get_schedule_ids(origin_conn, schedule_names)
    origin_schedule_details = get_schedule_details(origin_conn, origin_schedule_ids)
    base_schedules = create_base_schedules(destination_conn, origin_schedule_details, clone_name_prefix)
    base_schedule_names = [schedule['name'] for schedule in base_schedules]

    destination_schedule_ids = get_schedule_ids(destination_conn, base_schedule_names)
    destination_schedule_details = get_schedule_details(destination_conn, destination_schedule_ids)
    cloned_schedules = update_schedule_state(destination_conn, origin_schedule_details, destination_schedule_details)
    override_schedule_state(destination_conn, state_override, schedule_names=base_schedule_names)
    return cloned_schedules



from tableau_api_lib.api_requests import BaseRequest


class AddTagsRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding tags to content.
    :param class ts_connection: the Tableau Server connection object
    :param list tags: a list of tag names to add
    """
    def __init__(self,
                 ts_connection,
                 tags=None):
        super().__init__(ts_connection)
        self._tags = tags or []
        self.base_add_tags_request()

    def base_add_tags_request(self):
        self._request_body.update({'tags': {'tag': []}})
        return self._request_body

    def modified_add_tags_request(self):
        for tag in self._tags:
            self._request_body['tags']['tag'].append({
                'label': tag
            })
        return self._request_body

    def get_request(self):
        return self.modified_add_tags_request()



from tableau_api_lib.api_requests import BaseRequest


class CreateExtractsForWorkbookRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating extracts for a workbook.
    :param class ts_connection: the Tableau Server connection object
    :param bool extract_all_datasources_flag: True if extracting all datasources, False otherwise
    :param list datasource_ids: a list of datasources to create extracts from, if extract_all_datasources_flag is False
    """
    def __init__(self,
                 ts_connection,
                 extract_all_datasources_flag=True,
                 datasource_ids=None):
        super().__init__(ts_connection)
        self._extract_all_datasources_flag = extract_all_datasources_flag
        self._datasource_ids = datasource_ids or []
        self._validate_inputs()
        self.base_create_extracts_for_workbook_request()

    def _validate_inputs(self):
        valid = True
        if self._extract_all_datasources_flag is False and not any(self._datasource_ids):
            valid = False
        if self._extract_all_datasources_flag is True and self._datasource_ids:
            valid = False
        if not valid:
            raise self._invalid_parameter_exception()

    @property
    def optional_extract_param_keys(self):
        return ['datasource_ids']

    @property
    def optional_extract_param_values(self):
        return [self._datasource_ids]

    def base_create_extracts_for_workbook_request(self):
        self._request_body.update({'datasources': {'includeAll': self._extract_all_datasources_flag}})
        return self._request_body

    def modified_create_extracts_for_workbook_request(self):
        if not self._extract_all_datasources_flag and any(self._datasource_ids):
            self._request_body['datasources'].update({'datasource': []})
            for datasource_id in self._datasource_ids:
                self._request_body['datasources']['datasource'].append({
                    'id': datasource_id
                })
        return self._request_body

    def get_request(self):
        return self.modified_create_extracts_for_workbook_request()
